import dbus
from dbus.mainloop.glib import DBusGMainLoop
import gi
gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "3.0")

bus = dbus.SystemBus(mainloop=DBusGMainLoop())
snapper = dbus.Interface(bus.get_object('org.opensuse.Snapper',
                                        '/org/opensuse/Snapper'),
                         dbus_interface='org.opensuse.Snapper')



import pkg_resources, sys, signal
from gi.repository import Gtk, GLib, GdkPixbuf, Gio
from snappergui.mainWindow import SnapperGUI
from snappergui.propertiesDialog import propertiesDialog


def start_ui():
    app = Application()
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    exit_status = app.run(sys.argv)
    sys.exit(exit_status)


class Application(Gtk.Application):
    def __init__(self):
        Gtk.Application.__init__(self)
        GLib.set_application_name("SnapperGUI")
        GLib.set_prgname('snappergui')

        self.snappergui = None

    def build_app_menu(self):
        builder = Gtk.Builder()
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "ui/app-menu.ui"))
        menu = builder.get_object('app-menu')
        self.set_app_menu(menu)

        propertiesAction = Gio.SimpleAction.new('properties', None)
        propertiesAction.connect('activate', self.show_configs_properties)
        self.add_action(propertiesAction)

        aboutAction = Gio.SimpleAction.new('about', None)
        aboutAction.connect('activate', self.about)
        self.add_action(aboutAction)

        quitAction = Gio.SimpleAction.new('quit', None)
        quitAction.connect('activate', self.quit)
        self.add_action(quitAction)

    def show_configs_properties(self, action, param):
        dialog = propertiesDialog(self, self.snappergui.window)
        dialog.dialog.run()
        dialog.dialog.hide()

    def about(self, action, param):
        pass

    def quit(self, action=None, param=None):
        self.snappergui.window.destroy()

    def do_startup(self):
        Gtk.Application.do_startup(self)
        self.build_app_menu()

    def do_activate(self):
        if not self.snappergui:
            self.snappergui = SnapperGUI(self)
        self.snappergui.window.present()



from snappergui import snapper
import pkg_resources
from gi.repository import Gtk


class createConfig(object):
    """docstring for createConfig"""

    def __init__(self, parent):
        super(createConfig, self).__init__()
        builder = Gtk.Builder()
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "glade/createConfig.glade"))
        self.dialog = builder.get_object("createConfig")
        self.dialog.set_transient_for(parent)
        builder.connect_signals(self)

        self.name = ""
        self.subvolume = ""
        self.fstype = "btrfs"
        self.template = "default"

        builder.get_object("fsTypeCombo").set_active(0)

    def on_name_changed(self, widget):
        self.name = widget.get_chars(0, -1)

    def on_subvolume_changed(self, widget):
        self.subvolume = widget.get_chars(0, -1)

    def on_fstype_changed(self, widget):
        self.fstype = widget.get_active_text()

    def on_template_changed(self, widget):
        self.template = widget.get_chars(0, -1)

    def run(self):
        return self.dialog.run()

    def destroy(self):
        self.dialog.destroy()



from snappergui import snapper
import pkg_resources, subprocess, dbus
from snappergui.createSnapshot import createSnapshot
from snappergui.createConfig import createConfig
from snappergui.deleteDialog import deleteDialog
from snappergui.changesWindow import changesWindow
from snappergui.snapshotsView import snapshotsView
from gi.repository import Gtk
from time import strftime, localtime
from pwd import getpwuid


class SnapperGUI():
    """docstring for SnapperGUI"""

    def __init__(self, app):
        super(SnapperGUI, self).__init__()
        self.builder = Gtk.Builder()
        self.builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                                   "glade/mainWindow.glade"))
        self.statusbar = self.builder.get_object("statusbar")
        self.snapshotsTreeView = self.builder.get_object("snapstreeview")
        self.configsGroup = self.builder.get_object("configsGroup")
        self.window = self.builder.get_object("applicationwindow1")
        self.stack = self.builder.get_object("stack1")
        self.builder.connect_signals(self)

        self.window.set_application(app)

        self.configView = {}

        for config in snapper.ListConfigs():
            name = str(config[0])
            self.configView[name] = snapshotsView(name)
            self.stack.add_titled(self.configView[name].scrolledwindow, name, name)
            self.configView[name].selection.connect("changed",
                                                    self.on_snapshots_selection_changed)

        self.init_dbus_signal_handlers()
        self.window.show()

    def snapshot_columns(self, snapshot):
        if snapshot[3] == -1:
            date = "Now"
        else:
            date = strftime("%a %x %R", localtime(snapshot[3]))
        return [snapshot[0],
                snapshot[1],
                snapshot[2],
                date,
                getpwuid(snapshot[4])[0],
                snapshot[5],
                snapshot[6]]

    def get_current_config(self):
        return self.stack.get_visible_child_name()

    def on_stack_visible_child_changed(self, stack, property):
        self.update_controlls_and_userdatatreeview()

    def on_snapshots_selection_changed(self, selection):
        self.update_controlls_and_userdatatreeview()

    def update_controlls_and_userdatatreeview(self):
        config = self.get_current_config()
        userdatatreeview = self.builder.get_object("userdatatreeview")
        if config is not None:
            model, paths = self.configView[config].selection.get_selected_rows()
        if config is None or len(paths) == 0:
            self.builder.get_object("snapshotActions").set_sensitive(False)
            userdatatreeview.set_model(None)
        else:
            self.builder.get_object("snapshotActions").set_sensitive(True)

            if len(paths) == 1 and not model.iter_has_child(model.get_iter(paths[0])):
                self.builder.get_object("view-changes").set_sensitive(False)
            else:
                self.builder.get_object("view-changes").set_sensitive(True)

            try:
                snapshot_data = snapper.GetSnapshot(config, model[model.get_iter(paths[0])][0])
                userdata_liststore = Gtk.ListStore(str, str)
                for key, value in snapshot_data[7].items():
                    userdata_liststore.append([key, value])
                userdatatreeview.set_model(userdata_liststore)
            except dbus.exceptions.DBusException:
                pass

    def on_create_snapshot(self, widget):
        dialog = createSnapshot(self.window, self.get_current_config())
        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.OK:
            newSnapshot = snapper.CreateSingleSnapshot(dialog.config,
                                                       dialog.description,
                                                       dialog.cleanup,
                                                       dialog.userdata)
        elif response == Gtk.ResponseType.CANCEL:
            pass

    def on_create_config(self, widget):
        dialog = createConfig(self.window)
        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.OK:
            snapper.CreateConfig(dialog.name,
                                 dialog.subvolume,
                                 dialog.fstype,
                                 dialog.template)
        elif response == Gtk.ResponseType.CANCEL:
            pass

    def on_delete_snapshot(self, widget):
        config = self.get_current_config()
        selection = self.configView[config].selection
        (model, paths) = selection.get_selected_rows()
        snapshots = []
        for path in paths:
            treeiter = model.get_iter(path)
            # avoid duplicates because post snapshots are always added
            if model[treeiter][0] not in snapshots:
                snapshots.append(model[treeiter][0])
            # if snapshot has post add that to delete
            if model.iter_has_child(treeiter):
                child_treeiter = model.iter_children(treeiter)
                snapshots.append(model[child_treeiter][0])
        dialog = deleteDialog(self.window, config, snapshots)
        response = dialog.run()
        if response == Gtk.ResponseType.YES and len(dialog.to_delete) > 0:
            snapper.DeleteSnapshots(config, dialog.to_delete)

    def on_open_snapshot_folder(self, widget):
        config = self.get_current_config()
        selection = self.configView[config].selection
        model, paths = selection.get_selected_rows()
        for path in paths:
            treeiter = model.get_iter(path)
            mountpoint = snapper.GetMountPoint(config, model[treeiter][0])
            if model[treeiter][6] != '':
                snapper.MountSnapshot(config, model[treeiter][0], 'true')
            subprocess.Popen(['xdg-open', mountpoint])
            self.statusbar.push(True,
                                "The mount point for the snapshot %s from %s is %s" %
                                (model[treeiter][0], config, mountpoint))

    def on_viewchanges_clicked(self, widget):
        config = self.get_current_config()
        selection = self.configView[config].selection
        model, paths = selection.get_selected_rows()
        if len(paths) > 1:
            # open a changes window with the first and the last snapshot selected
            begin = model[paths[0]][0]
            end = model[paths[-1]][0]
            window = changesWindow(config, begin, end)
        elif len(paths) == 1 and model.iter_has_child(model.get_iter(paths[0])):
            # open a changes window with the selected pre snapshot and its corresponding post snapshot
            child_iter = model.iter_children(model.get_iter(paths[0]))
            begin = model[paths[0]][0]
            end = model.get_value(child_iter, 0)
            window = changesWindow(config, begin, end)

    def init_dbus_signal_handlers(self):
        signals = {
            "SnapshotCreated": self.on_dbus_snapshot_created,
            "SnapshotModified": self.on_dbus_snapshot_modified,
            "SnapshotsDeleted": self.on_dbus_snapshots_deleted,
            "ConfigCreated": self.on_dbus_config_created,
            "ConfigModified": self.on_dbus_config_modified,
            "ConfigDeleted": self.on_dbus_config_deleted
        }
        for signal, handler in signals.items():
            snapper.connect_to_signal(signal, handler)

    def on_dbus_snapshot_created(self, config, snapshot):
        self.statusbar.push(True, "Snapshot %s created for %s" % (str(snapshot), config))
        self.configView[config].add_snapshot_to_tree(str(snapshot))

    def on_dbus_snapshot_modified(self, config, snapshot):
        print("Snapshot SnapshotModified")

    def on_dbus_snapshots_deleted(self, config, snapshots):
        snaps_str = ""
        for snapshot in snapshots:
            snaps_str += str(snapshot) + " "
        self.statusbar.push(True, "Snapshots deleted from %s: %s" % (config, snaps_str))
        for deleted in snapshots:
            self.configView[config].remove_snapshot_from_tree(deleted)

    def on_dbus_config_created(self, config):
        self.configView[config] = snapshotsView(config)
        self.configView[config].update_view()
        self.stack.add_titled(self.configView[config]._TreeView, config, config)
        self.statusbar.push(5, "Created new configuration %s" % config)

    def on_dbus_config_modified(self, args):
        print("Config Modified")

    def on_dbus_config_deleted(self, args):
        print("Config Deleted")

    def on_main_destroy(self, args):
        for config in snapper.ListConfigs():
            for snapshot in snapper.ListSnapshots(config[0]):
                if snapshot[6] != '':
                    snapper.UmountSnapshot(config[0], snapshot[0], 'true')



from snappergui import snapper
import pkg_resources
from gi.repository import Gtk
from pwd import getpwuid


class deleteDialog(object):
    """docstring for deleteDialog"""

    def __init__(self, parent, config, snapshots):
        super(deleteDialog, self).__init__()
        builder = Gtk.Builder()
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "glade/deleteSnapshot.glade"))

        self.dialog = builder.get_object("dialogDelete")
        self.dialog.set_transient_for(parent)
        self.deletetreeview = builder.get_object("deletetreeview")
        builder.connect_signals(self)

        self.snapshots_list = snapshots
        self.to_delete = snapshots

        parents = []
        self.deleteTreeStore = Gtk.TreeStore(bool, int, str, str)
        for snapshot in snapshots:
            snapinfo = snapper.GetSnapshot(config, snapshot)
            # self.deleteTreeStore.append(self.get_row(snapinfo))
            if snapinfo[1] == 1:  # Pre snapshot
                parents.append(self.deleteTreeStore.append(None, self.get_row(snapinfo)))
            elif snapinfo[1] == 2:  # Post snapshot
                parent_node = None
                for parent in parents:
                    if self.deleteTreeStore.get_value(parent, 1) == snapinfo[2]:
                        parent_node = parent
                        break
                self.deleteTreeStore.append(parent_node)
            else:  # Single snapshot
                self.deleteTreeStore.append(None, self.get_row(snapinfo))
        self.deletetreeview.set_model(self.deleteTreeStore)
        self.deletetreeview.expand_all()

    def get_row(self, snapshot):
        return [True, snapshot[0], getpwuid(snapshot[4])[0], snapshot[5]]

    def run(self):
        response = self.dialog.run()
        self.dialog.destroy()
        return response

    def on_toggle_delete_snapshot(self, widget, path):
        treeiter = self.deleteTreeStore.get_iter(path)
        self.deleteTreeStore.set_value(treeiter,
                                       False,
                                       not (self.deleteTreeStore.get_value(treeiter, False)))
        snapshot_num = self.deleteTreeStore.get_value(treeiter, True)
        if self.deleteTreeStore.get_value(treeiter, False):
            if snapshot_num not in self.to_delete:
                self.to_delete.append(snapshot_num)
        else:
            if snapshot_num in self.to_delete:
                self.to_delete.remove(snapshot_num)



from snappergui import snapper
import pkg_resources
from gi.repository import Gtk, Gdk


class createSnapshot(object):
    TYPE_HERE = "<Type here>"

    """docstring for createSnapshot"""
    def __init__(self, parent, config_name):
        super(createSnapshot, self).__init__()
        builder = Gtk.Builder()
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "glade/createSnapshot.glade"))

        self.dialog = builder.get_object("dialogCreate")
        self.userdataTree = builder.get_object("userdatatreeview")
        self.userdataTree.get_model().append([self.TYPE_HERE, ""])
        self.dialog.set_transient_for(parent)
        builder.connect_signals(self)

        self.config = ""
        self.description = ""
        self.cleanup = ""
        self.userdata = {}

        configscombo = Gtk.ListStore(str)
        combobox = builder.get_object("configsCombo")
        combobox.set_model(configscombo)
        for i, config in enumerate(snapper.ListConfigs()):
            name = str(config[0])
            configscombo.append([name])
            if name == config_name:
                combobox.set_active(i)

        builder.get_object("cleanupcombo").set_active(0)

    def on_config_changed(self,widget):
        self.config = widget.get_model()[widget.get_active()][0]

    def on_description_changed(self,widget):
        self.description = widget.get_chars(0,-1)

    def on_cleanup_changed(self,widget):
        self.cleanup = widget.get_active_text()
        if self.cleanup == "None":
            self.cleanup = ""

    def _on_key_press(self, widget, event):
        if event.keyval == Gdk.KEY_Delete:
            model, selected = self.userdataTree.get_selection().get_selected_rows()
            if str(selected[0]) != str(len(model) - 1):
                del model[selected[0]]

    def _on_editing_started(self, cell, editable, path):
        if editable.get_text() == self.TYPE_HERE:
            editable.set_text("")

    def _on_name_edited(self, renderer, path, new_text):
        userdatamodel = self.userdataTree.get_model()
        if userdatamodel[path][0] == self.TYPE_HERE:
            userdatamodel.append([self.TYPE_HERE, ""])
        if new_text == "" or new_text == self.TYPE_HERE:
            del userdatamodel[path]
        else:
            userdatamodel[path][0] = new_text

    def _on_value_edited(self, renderer, path, new_text):
        userdatamodel = self.userdataTree.get_model()
        userdatamodel[path][1] = new_text

    def get_userdata_from_model(self, model):
        for row in model:
            if row[0] != self.TYPE_HERE:
                self.userdata[row[0]] = row[1]

    def run(self):
        response = self.dialog.run()
        self.get_userdata_from_model( self.userdataTree.get_model() )
        return response

    def destroy(self):
        self.dialog.destroy()



from snappergui import snapper
import pkg_resources, dbus
from gi.repository import Gtk


class PropertiesTab(object):
    """docstring for PropertiesTab"""

    def __init__(self, config):
        builder = Gtk.Builder()
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "glade/propertiesDialog.glade"))
        self.configsGrid = builder.get_object("configsGrid")

        self.widgets = {}
        for k, v in config[2].items():
            widget = builder.get_object(k)
            # Values are set here depending on their types
            if type(widget) == Gtk.Entry:
                widget.set_text(v)
            elif type(widget) == Gtk.SpinButton:
                adjustment = Gtk.Adjustment(value=int(v),
                                            lower=0, upper=5000,
                                            step_increment=1,
                                            page_increment=10,
                                            page_size=0)
                widget.set_adjustment(adjustment)
            elif type(widget) == Gtk.Switch:
                if v == "yes":
                    widget.set_active(True)
                elif v == "no":
                    widget.set_active(False)
            else:
                print("ERROR: Could not handle property \"%s\"." % k)
            self.widgets[k] = widget

    def get_current_value(self, setting):
        widget = self.widgets[setting]
        if type(widget) == Gtk.Entry:
            return widget.get_text()
        elif type(widget) == Gtk.Switch:
            if widget.get_active():
                return "yes"
            else:
                return "no"
        elif type(widget) == Gtk.SpinButton:
            return str(int(widget.get_value()))


class propertiesDialog(object):
    """docstring for propertiesDialog"""

    def __init__(self, widget, parent):
        self.parent = parent

        builder = Gtk.Builder()
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "glade/propertiesDialog.glade"))
        self.dialog = builder.get_object("dialogProperties")
        self.notebook = builder.get_object("notebookProperties")
        builder.connect_signals(self)

        self.dialog.set_transient_for(parent)

        self.tabs = {}
        for config in snapper.ListConfigs():
            currentTab = PropertiesTab(config)
            self.tabs[str(config[0])] = currentTab
            self.notebook.append_page(currentTab.configsGrid, Gtk.Label.new(config[0]))
        self.notebook.show_all()

    def get_changed_settings(self, config):
        changed = {}
        for k, v in snapper.GetConfig(config)[2].items():
            currentValue = self.tabs[config].get_current_value(k)
            if currentValue and v != currentValue:
                changed[k] = currentValue
        return changed

    def on_response(self, widget, response):
        if response == Gtk.ResponseType.OK:
            currentConfig = str(snapper.ListConfigs()[self.notebook.get_current_page()][0])
            try:
                snapper.SetConfig(currentConfig, self.get_changed_settings(currentConfig))
            except dbus.exceptions.DBusException as error:
                if str(error).find("error.no_permission") != -1:
                    self.dialog.destroy()
                    dialog = Gtk.MessageDialog(self.parent, 0, Gtk.MessageType.WARNING,
                                               Gtk.ButtonsType.OK,
                                               "You don't have permission to edit configurations")
                    dialog.run()
                    dialog.destroy()



from snappergui import snapper
import pkg_resources, dbus
from gi.repository import Gtk
from time import strftime, localtime
from pwd import getpwuid


class snapshotsView(Gtk.Widget):
    """docstring for snapshotsView"""

    def __init__(self, config):
        super(snapshotsView, self).__init__()
        self.config = config
        self.builder = Gtk.Builder()
        self.builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                                   "glade/snapshotsView.glade"))
        self.builder.connect_signals(self)
        self._TreeView = self.builder.get_object("snapstreeview")
        self.selection = self.builder.get_object("snapshotsSelection")
        self.scrolledwindow = self.builder.get_object("scrolledwindow1")
        self.count = 0
        self._TreeView.connect("realize", self.update_view)

    def update_view(self, widget=None, user_data=None):
        treestore = self.get_config_treestore()
        if treestore:
            self._TreeView.set_model(treestore)

    def snapshot_columns(self, snapshot):
        if snapshot[3] == -1:
            date = "Now"
        else:
            date = strftime("%a %x %R", localtime(snapshot[3]))
        return [snapshot[0],
                snapshot[1],
                snapshot[2],
                date,
                getpwuid(snapshot[4])[0],
                snapshot[5],
                snapshot[6]]

    def get_config_treestore(self):
        configstree = Gtk.TreeStore(int, int, int, str, str, str, str)
        # Check if the user has permission to see/edit this configuration
        try:
            snapshots_list = snapper.ListSnapshots(self.config)
        except dbus.exceptions.DBusException:
            return
        parents = []
        self.count = len(snapshots_list)
        for snapshot in snapshots_list:
            if snapshot[1] == 1:  # Pre snapshot
                parents.append(configstree.append(None, self.snapshot_columns(snapshot)))
            elif snapshot[1] == 2:  # Post snapshot
                parent_node = None
                for parent in parents:
                    if configstree.get_value(parent, 0) == snapshot[2]:
                        parent_node = parent
                        break
                configstree.append(parent_node, self.snapshot_columns(snapshot))
            else:  # Single snapshot
                configstree.append(None, self.snapshot_columns(snapshot))
        return configstree

    def add_snapshot_to_tree(self, snapshot, pre_snapshot=None):
        treemodel = self._TreeView.get_model()
        # Ignore if we are notified about snapshot we can't read details about.
        try:
            snapinfo = snapper.GetSnapshot(self.config, snapshot)
        except dbus.exceptions.DBusException:
            return
        pre_number = snapinfo[2]
        if snapinfo[1] == 2:  # if type is post
            for aux, row in enumerate(treemodel):
                if pre_number == row[0]:
                    pre_snapshot = treemodel.get_iter(aux)
                    break
        treemodel.append(pre_snapshot, self.snapshot_columns(snapinfo))

    def remove_snapshot_from_tree(self, snapshot):
        treemodel = self._TreeView.get_model()
        for aux, row in enumerate(treemodel):
            if snapshot == row[0]:
                has_child = treemodel.iter_has_child(treemodel.get_iter(aux))
                if has_child:
                    # FIXME meh
                    self.update_view()
                else:
                    del treemodel[aux]

    def on_description_edited(self, widget, treepath, text):
        snapshot_row = self._TreeView.get_model()[treepath]
        snapshot_num = snapshot_row[0]
        snapshot_info = snapper.GetSnapshot(self.config, snapshot_num)
        snapper.SetSnapshot(self.config, snapshot_info[0], text, snapshot_info[6], snapshot_info[7])
        snapshot_row[5] = text

    def on_cleanup_edited(self, widget, treepath, text):
        snapshot_row = self._TreeView.get_model()[treepath]
        snapshot_num = snapshot_row[0]
        snapshot_info = snapper.GetSnapshot(self.config, snapshot_num)
        snapper.SetSnapshot(self.config, snapshot_info[0], snapshot_info[5], text, snapshot_info[7])
        snapshot_row[6] = text



from snappergui import snapper
import pkg_resources, os, time, difflib, collections
from gi.repository import Gtk, GtkSource, GObject, Gdk


class StatusFlags(object):
    """File status flags from https://github.com/openSUSE/snapper/blob/master/snapper/File.h#L39-L51"""

    # @formatter:off
    CREATED     =   1  # created
    DELETED     =   2  # deleted
    TYPE        =   4  # type has changed
    CONTENT     =   8  # content has changed
    PERMISSIONS =  16  # permissions have changed, see chmod(2)
    OWNER       =  32  # owner has changed, see chown(2)
    USER        =  32  # deprecated - alias for OWNER
    GROUP       =  64  # group has changed, see chown(2)
    XATTRS      = 128  # extended attributes changed, see attr(5)
    ACL         = 256  # access control list changed, see acl(5)
    # @formatter:on


class changesWindow(object):
    """docstring for changesWindow"""

    TreeNode = collections.namedtuple("TreeNode", "path, children, status, is_dir")

    def __init__(self, config, begin, end):
        super(changesWindow, self).__init__()
        builder = Gtk.Builder()
        GObject.type_register(GtkSource.View)
        builder.add_from_file(pkg_resources.resource_filename("snappergui",
                                                              "glade/changesWindow.glade"))

        builder.get_object("titlelabel").set_text("%s -> %s" % (begin, end))
        self.window = builder.get_object("changesWindow")
        self.statusbar = builder.get_object("statusbar1")
        self.pathstreeview = builder.get_object("pathstreeview")
        self.fileview = builder.get_object("fileview")
        self.choicesviewgroup = builder.get_object("actiongroup")
        builder.connect_signals(self)

        # save mountpoints for begin and end snapshots
        self.beginpath = snapper.GetMountPoint(config, begin)
        self.endpath = snapper.GetMountPoint(config, end)
        self.config = config
        self.snapshot_begin = begin
        self.snapshot_end = end

        self.choicesviewgroup.get_action("begin").set_label(str(begin))
        self.choicesviewgroup.get_action("end").set_label(str(end))

        self.sourcebuffer = GtkSource.Buffer()
        self.fileview.set_buffer(self.sourcebuffer)

        self.window.show_all()
        GObject.idle_add(self.on_idle_init_paths_tree)

    def add_path_to_tree(self, path, status, tree):
        is_dir = os.path.isdir(self.beginpath + path) or os.path.isdir(self.endpath + path)
        parts = path.split('/')
        node = tree
        # Add directories to tree
        for file_name in parts[1:-1]:
            if not file_name in node.children:
                node.children[file_name] = changesWindow.TreeNode("", {}, 0, True)
            node = node.children[file_name]
        # Add last part of path to tree
        if is_dir:
            node.children[parts[-1]] = changesWindow.TreeNode("", {}, status, True)
        else:
            node.children[parts[-1]] = changesWindow.TreeNode(path, None, status, False)

    def print_tree(self, tree, indent=""):
        for name, child in tree.children.items():
            print(indent + name + ("/" if child.is_dir else "") + "   (" +
                  self.file_status_to_string(child.status) + ")")
            if child.is_dir:
                self.print_tree(child, indent + "    ")

    def file_status_to_string(self, status):
        if status & StatusFlags.CREATED:  # Created file
            return "Created"
        elif status & StatusFlags.DELETED:  # Deleted file
            return "Deleted"
        elif status > 0:  # Modified file
            modification_types = {
                StatusFlags.ACL: "acl",
                StatusFlags.CONTENT: "content",
                StatusFlags.GROUP: "group",
                StatusFlags.OWNER: "owner",
                StatusFlags.XATTRS: "xattrs",
                StatusFlags.TYPE: "inode type",
                StatusFlags.PERMISSIONS: "permissions"
            }
            modified_list = []
            for flag, name in modification_types.items():
                if status & flag:
                    modified_list.append(name)
            return "Modified: " + ", ".join(modified_list)
        else:
            return "No changes"

    def get_treestore_from_tree(self, tree):
        # Row: [gtk-stock-icon, file name, file complete path, entry color, tooltip]
        treestore = Gtk.TreeStore(str, str, str, Gdk.RGBA, str)

        def get_children(subtree, parent=None):
            for file_name, child in subtree.children.items():
                color = Gdk.RGBA(0.0, 0.0, 0.0)
                if child.status & StatusFlags.CREATED:  # Created file
                    color = Gdk.RGBA(0.0, 0.57, 0.0)
                elif child.status & StatusFlags.DELETED:  # Deleted file
                    color = Gdk.RGBA(0.6, 0.0, 0.0)
                elif child.status > 0:  # Modified file
                    color = Gdk.RGBA(0.49, 0.47, 0.0)

                node = treestore.append(parent, [
                    Gtk.STOCK_DIRECTORY if child.is_dir else Gtk.STOCK_FILE,
                    file_name, child.path, color, self.file_status_to_string(child.status)
                ])
                # if this child is a directory get childs
                if child.children is not None:
                    get_children(child, node)

        get_children(tree)
        return treestore

    def on_query_tooltip(self, widget, x, y, keyboard_tip, tooltip):
        if not widget.get_tooltip_context(x, y, keyboard_tip):
            return False
        else:
            on_row, x, y, model, path, it = widget.get_tooltip_context(x, y, keyboard_tip)
            if on_row:
                tooltip.set_text(model[it][4])
                widget.set_tooltip_row(tooltip, path)
                return True
            else:
                return False

    def on_idle_init_paths_tree(self):
        snapper.CreateComparison(self.config, self.snapshot_begin, self.snapshot_end)

        dbus_array = snapper.GetFiles(self.config, self.snapshot_begin, self.snapshot_end)

        # create structure to sort paths into tree
        files_tree = changesWindow.TreeNode("/", {}, 0, True)
        for entry in dbus_array:
            self.add_path_to_tree(str(entry[0]), int(entry[1]), files_tree)

        # self.print_tree(files_tree)
        self.pathstreeview.set_model(self.get_treestore_from_tree(files_tree))
        # self.pathstreeview.expand_all()

        # display in statusbar how many files have changed
        self.statusbar.push(1, "%d files" % len(dbus_array))

        snapper.DeleteComparison(self.config, self.snapshot_begin, self.snapshot_end)

        # we dont want this function to be called anymore
        return False

    def get_lines_from_file(self, path):
        try:
            return open(path).readlines()
        except IsADirectoryError:
            pass
        except FileNotFoundError:
            return ""
        except UnicodeDecodeError:
            pass
        except PermissionError:
            print("PermissionError")
            pass  # TODO maybe display a dialog with the error?
        return None

    def _on_pathstree_selection_changed(self, selection):
        (model, treeiter) = selection.get_selected()
        if treeiter is not None and model[treeiter] != "":
            # append file path to snapshot mountpoint
            fromfile = self.beginpath + model[treeiter][2]
            tofile = self.endpath + model[treeiter][2]

            fromlines = self.get_lines_from_file(fromfile)
            if fromlines is None:
                return
            elif fromlines == "":
                fromfile = "New File"
                fromdate = ""
            else:
                fromdate = time.ctime(os.stat(fromfile).st_mtime)

            tolines = self.get_lines_from_file(tofile)
            if tolines is None:
                return
            elif tolines == "":
                tofile = "Deleted File"
                todate = ""
            else:
                todate = time.ctime(os.stat(tofile).st_mtime)

            languagemanager = GtkSource.LanguageManager()
            currentview = self.choicesviewgroup.get_action("end").get_current_value()

            if currentview == 0:  # show file from begin snapshot
                self.sourcebuffer.set_language(languagemanager.get_language("text"))
                self.sourcebuffer.set_text("".join(fromlines))
            elif currentview == 1:  # show diff of file changes between snapshots
                self.sourcebuffer.set_language(languagemanager.get_language("diff"))
                difflines = difflib.unified_diff(fromlines,
                                                 tolines,
                                                 fromfile=fromfile,
                                                 tofile=tofile,
                                                 fromfiledate=fromdate,
                                                 tofiledate=todate)
                self.sourcebuffer.set_text("".join(difflines))
            elif currentview == 2:  # show file from end snapshot
                self.sourcebuffer.set_language(languagemanager.get_language("text"))
                self.sourcebuffer.set_text("".join(tolines))
